/*--------------------------------------------------------------------------
| ftty.cpp - Simple fossil/bios TTY program to test serial-ports.
Copyright 1995-97 Comtrol Corp.
 6-18-97 - fix sending file option, add flow control settings.
|--------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>

#define getdosticks() (*((unsigned int far *)(0x46cL)))

#define WANT_BASE_TYPES

#ifdef WANT_BASE_TYPES
typedef unsigned char  BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef int BOOL;
#endif

#include "fu.c"   // fossil functions, still allow easy one module compile.

typedef struct {
  int port;
  DWORD BaudRate;
  int Parity;
  int StopBits;
  int ByteSize;
  int flow_ctrl; // 1 = rts,cts 2 = xon,xoff
} DCB;

static int show_line_status(WORD stat);
static int show_fossil(void);
static int show_status(void);
static int parse_cmdline(char *buf, char *port_name, DCB *dcb);
int f_set_com_options(int port,
                      DWORD BaudRate,
                      int Parity,
                      int StopBits,
                      int ByteSize,
                      int flow_ctrl);

// Com stuff
int port_num;
int use_fossil = 1;
char port_name[200];
char tmpstr[200];
DCB setup_dcb;  // our cmd line settings
char flow_control = 0;

unsigned char buf[600];
unsigned char send_buf[100];
DWORD ModemStatus;
DWORD OldModemStatus;
char file_name[80];
FossInfo fi;
FILE *fp_send = NULL;

/*--------------------------------------------------------------------------
| main -
|--------------------------------------------------------------------------*/
int main(int argc, char **argv)
{
 unsigned char c;
 WORD stat, i;
 DCB our_dcb;
 DCB *dcb;
 char par_char, stop_char;
 DWORD ret_stat;
 DWORD cnt;
 int in_cnt;
 int out_free;
  file_name[0] = 0;
  port_name[0] = 0;
  memset(&setup_dcb, 0, sizeof(setup_dcb));
  setup_dcb.StopBits = 0;
  setup_dcb.BaudRate = 9600;
  setup_dcb.ByteSize = 8;
  setup_dcb.Parity = 0;
  setup_dcb.flow_ctrl = 0;  // none

  if (argc == 1)
  {
    printf("EXAMPLE>ftty COM5      (defaults to 9600,8,n,1) \n");
    printf("EXAMPLE>ftty COM3:1200\n");
    printf("EXAMPLE>ftty COM7:9600,8,n,1 \n");
    printf("EXAMPLE>ftty COM7:9600,8,n,1,r filesend.txt\n");
    printf("  parameters include: BAUD,databits,parity,stopbits,flowcontrol\n");
    printf("  where flowcontrol=X for xon,xoff, R=rts,cts, N=none\n");
#ifdef COMMENT_OUT
    return 1;
#endif

    printf("Enter COM port name to open(example:COM5):");
    scanf("%s", port_name);
    if (strlen(port_name) < 2)
    {
      printf("Bad Port Name\n");
      return 1;
    }
  }
  else
  {
    stat = parse_cmdline(argv[1], port_name, &setup_dcb);

    if (argc == 3)
    {
      strcpy(file_name, argv[2]);
      printf("sending file %s\n", file_name);
    }
  }

                         // get the index of the port from the port name
  if (isdigit(port_name[0]))
    port_num = atoi(port_name);
  else
  {
    i = 0;
    while ( (!isdigit(port_name[i])) && (port_name[i] != 0) )
      ++i;
    if (port_name[i] == 0)
    {
      printf("Bad Port Name\n");
      exit(1);
    }
    port_num = atoi(&port_name[i]);
  }
  if (port_num == 0)
  {
    printf("Use port 1-indexed port numbers: COM1 = first com port.\n");
    exit(1);
  }
  if (port_num > 255)
  {
    printf("Port number out of range.\n");
    exit(1);
  }
  --port_num;  // fossil port numbers are 0-indexed.

  if (use_fossil)
  {
    stat = FOSSinit(port_num);
    if (stat != 0x1954)
    {
      printf("Could not open FOSSIL Port: COM%d.\n", port_num+1);
      printf("Will use BIOS calls.\n");
      use_fossil = 0;
    }
    else printf("Fossil Driver Present\n");
  }

  if (setup_dcb.BaudRate == 0)
    setup_dcb.BaudRate = 9600;

  if ((setup_dcb.ByteSize != 7) || (setup_dcb.ByteSize != 8))
    setup_dcb.ByteSize = 8;

  switch(setup_dcb.Parity)
  {
    case 0:  par_char = 'n'; break; // none
    case 1:  par_char = 'o'; break;
    case 2:  par_char = 'e'; break;
    case 3:  par_char = 'm'; break;
    case 4:  par_char = 's'; break;
  }

  switch (setup_dcb.StopBits)
  {
    case 0: stop_char = '1'; break;
    case 2: stop_char = '2'; break;
  }

  printf("Using %s:%ld,%d,%c,%c... (Esc to Exit)\n",
      port_name,
      setup_dcb.BaudRate,
      setup_dcb.ByteSize,
      par_char,
      stop_char);

  dcb->StopBits = setup_dcb.StopBits;
  dcb->BaudRate = setup_dcb.BaudRate;
  dcb->ByteSize = setup_dcb.ByteSize;
  dcb->Parity   = setup_dcb.Parity;

  if (use_fossil)
  {
    setup_dcb.flow_ctrl = 0; // 0=none, 1 = rts,cts, 2 = xon,xoff
    switch(flow_control)
    {
      case 'X': 
         printf("xon,xoff used\n");
         setup_dcb.flow_ctrl = 2; // 0=none, 1 = rts,cts, 2 = xon,xoff
      break;
      case 'R': 
         printf("rts,cts handshake used\n");
         setup_dcb.flow_ctrl = 1; // 0=none, 1 = rts,cts, 2 = xon,xoff
      break;
      case 'N': 
         printf("no handshake used\n");
         setup_dcb.flow_ctrl = 0; // 0=none, 1 = rts,cts, 2 = xon,xoff
      break;
    }

    stat = f_set_com_options(port_num,
                      setup_dcb.BaudRate,
                      setup_dcb.Parity,
                      setup_dcb.StopBits,
                      setup_dcb.ByteSize,
                      setup_dcb.flow_ctrl); // 0=none, 1 = rts,cts, 2 = xon,xoff

    FOSSgetinfo(port_num, &fi);
    printf("Fossil Driver %s\n Version: %d.%d\n",fi.ident, fi.majver,fi.minver);
  }
  else
  {
    BIOSconfig(port_num, 
                      setup_dcb.BaudRate,
                      setup_dcb.ByteSize,
                      setup_dcb.Parity,
                      setup_dcb.StopBits);
  }

  // if filename given at end of complete option list, then open file
  // send data out the port, and exit.
  if (file_name[0] != 0)
  {
    printf("Sending file: %s\n",file_name);
    fp_send = fopen(file_name, "rb");
    if (fp_send == NULL)
    {
      printf("bad filename\n");
    }
  }


  // otherwise do normal tty stuff: display chars coming in, send out
  // any chars hit on the keyboard.

  c = 0;
  printf("Press ESC to exit(CAPS-LOCK must be off with Win95.)\n");
  for (;;)
  {
    while (!kbhit())
    {
      if (fp_send)  // if we are sending a file..
      {
        if (use_fossil)
        {
          FOSSgetinfo(port_num, &fi);
          if (fi.ofree > 80)
          {
            i = fread(send_buf,1,80,fp_send);
            if (i != 0)
              FOSSwriteblk(port_num, i, send_buf);
            else
            {
              printf("Done sending file.");
              fp_send = NULL;
            }
          }
        }
        else  // just use BIOS
        {
          if (BIOSstatus(port_num) & LNSTAT_THRE)
          {
            i = fread(send_buf,1,1,fp_send);
    
            if (i != 0)
              BIOSsendchar(port_num, send_buf[0]);
            else
            {
              printf("Done sending file.");
              fp_send = NULL;
            }
          }
        }
      } // fp_send

      // check status, and check for keyboard entry
      if (use_fossil)
      {
        FOSSgetinfo(port_num, &fi);
        in_cnt  =  (fi.ibufr - fi.ifree);
        // out_free = fi.ofree;
        if (in_cnt)
        {
          if (in_cnt > 80) in_cnt = 80; // limit
          cnt = FOSSreadblk(port_num, in_cnt, buf);
          buf[cnt] = 0;  // null terminate.
          printf("%s",buf);  // print rec. data to screen
        }
      }
      else // bios mode
      {
        if (BIOSstatus(port_num) & LNSTAT_DATARDY)
        {
          buf[0] = BIOSgetchar(port_num) & 0xff;
          printf("%c",buf[0]);  // print rec. data to screen
        }
      }
    }

    c = getch();    // grab keyboard char.

    if (use_fossil)
    {
      if (c == 0x1b)
      {
        printf("====================================================\n");
        printf("Menu\n");
        printf("ESC-Back to Terminal Mode.\n");
        printf("A-Ascii send file\n");
        printf("X-eXit Program\n");
        printf("F-Fossil Info\n");
        printf("S-Fossil Port Status\n");
        printf("d- Clear DTR  D- Set DTR  r- Clear RTS  R- Set RTS\n");
        printf("====================================================\n");

        c = getch();    // grab keyboard char.
 
        printf("\n");
        for (;;)
        {
          if (c == 0x1b)
            break;
          if ((c == 'd') ||
              (c == 'D') ||
              (c == 'r') ||
              (c == 'R'))
          {
            if (c == 'd') FOSSdtr(port_num, 0);
            if (c == 'D') FOSSdtr(port_num, 1);
            if (c == 'r') FOSSrts(port_num, 0);
            if (c == 'R') FOSSrts(port_num, 1);
  
            printf("==");
            stat = FOSSstatus(port_num);
            show_line_status(stat);
            printf("\n");
          }
  
          c = toupper(c);  // kill case
  
          if (c == 'F')
          {
            show_fossil();
          }
  
          if (c == 'A')  // ascii send file
          {
            printf("File to send:");
            scanf("%s", file_name);
            printf("\n");

            fp_send = fopen(file_name, "rb");
            if (fp_send == NULL)
              printf("bad filename\n");
            else printf("Sending file: %s\n",file_name);
            break;  // back to terminal mode
          }

          if (c == 'S')
          {
            show_status();
            printf("==");
          stat = FOSSstatus(port_num);
            show_line_status(stat);
          printf("\n");
          }
          if (c == 'X')
          {
            FOSSclose(port_num);
          exit(0);
          }
          c = getch();    // grab keyboard char.
          printf("\n");
        }  // for (;;)
        printf("====Back to Terminal Mode:\n");
      }
      else
        FOSSwriteblk(port_num, 1, &c);
    }  
    else  // else bios mode
    {
      if (c == 0x1b)
        break;
      else
      {
        BIOSsendchar(port_num, c);
      }
    }
  } // for (;;)
  
  //  FOSSclose(port_num);
  //  return(0);
  return 0;
}
  
  /*--------------------------------------------------------------------
  | show_line_status -
  |---------------------------------------------------------------------*/
  static int show_line_status(WORD stat)
  {
   BYTE bstat;
 
  printf("Inputs[CTS:");
  if (stat & MDMSTAT_CTS)
       printf("ON ");
  else printf("OFF");

  printf(" DSR:");
  if (stat & MDMSTAT_DSR)
       printf("ON ");
  else printf("OFF");

  printf(" CD:");
  if (stat & MDMSTAT_CD)
       printf("ON ");
  else printf("OFF");

  printf(" RI:");
  if (stat & MDMSTAT_RI)
       printf("ON ");
  else printf("OFF");
  printf("] ");

// BL = Modem control register, where:
//      Bits 7-5 = 0 (Reserved)
//      Bit 4    = 1 Loopback mode
//      Bit 3    = 1 OUT2 (interrupts) enabled
//      Bit 2    = 1 OUT1 active
//      Bit 1    = 1 Request to send active
//      Bit 0    = 1 Data terminal ready (DTR) active

  bstat = FOSSgetMCR(port_num);
  printf("  Outputs[DTR:");
  if (bstat & 0x1)
       printf("ON ");
  else printf("OFF");

  printf(" RTS:");
  if (bstat & 0x2)
       printf("ON ");
  else printf("OFF");

  printf("]");

  printf(" LOOPBACK:");
  if (bstat & 0x10)
       printf("ON ");
  else printf("OFF");

  return 0;
}

/*--------------------------------------------------------------------
| show_fossil -
|---------------------------------------------------------------------*/
static int show_fossil(void)
{
  FOSSgetinfo(port_num, &fi);

  printf("Fossil Driver[%s]\n", fi.ident);
  printf("Version:%d.%d\n",fi.majver,fi.minver);
  printf("ID structure size:%d\n",fi.strsiz);
  printf("Baud:%d\n",fi.baud);
  printf("Output Buffer Size:%d Free space:%d\n",fi.obufr,fi.ofree);
  printf(" Input Buffer Size:%d Free space:%d\n",fi.ibufr, fi.ifree);
  printf("Screen Width:%d Height:%d\n",fi.swidth, fi.sheight);
  return 0;
}

/*--------------------------------------------------------------------
| show_status -
|---------------------------------------------------------------------*/
static int show_status(void)
{
 static char *stat_strings[] = {
 "MDMSTAT_DCTS  ",
 "MDMSTAT_DDSR  ",
 "MDMSTAT_DRI   ",
 "MDMSTAT_DCD   ",
 "MDMSTAT_CTS   ",
 "MDMSTAT_DSR   ",
 "MDMSTAT_RI    ",
 "MDMSTAT_CD    ",
 "LNSTAT_DATARDY",
 "LNSTAT_OVRRUN ",
 "LNSTAT_PARITY ",
 "LNSTAT_FRAMING",
 "LNSTAT_BREAK  ",
 "LNSTAT_THRE   ",
 "LNSTAT_TE     ",
 "LNSTAT_TIMEOUT", NULL};
 
 WORD stat, bit,i;

  bit = 1;
  stat = FOSSstatus(port_num);
  for (i=0; i<16; i++)
  {
    printf("%2d-%2xH) %s",i,i, stat_strings[i]);
    if (stat & bit)
         printf(" ON\n");
    else printf(" OFF\n");
    bit <<= 1;
  }
  return 0;
}

/*--------------------------------------------------------------------
| parse_cmdline - Grab info in format "COM#:9600,8,n,1".
|---------------------------------------------------------------------*/
static int parse_cmdline(char *buf, char *port_name, DCB *dcb)
{
 int i;

  for (i=0; i<10; i++)
  {
    port_name[i] = buf[i];
    if ((buf[i] == ':') || (buf[i] == 0) || (buf[i] == ','))
    {
      port_name[i] = 0;
      break;
    }
  }

  if (buf[i] != ':') return 1;

  while ((*buf != 0) && (*buf != ':'))  // pass up COM#:
    ++buf;
  if (*buf == 0) return 0;  // end after COM#
  ++buf;
  if (!((*buf >= '0') && (*buf <= '9')))
  {
    printf("bad baudrate option\n");
    return 0;
  }

  dcb->BaudRate = atol(buf);

  printf("baud:%ld", dcb->BaudRate);

  while ((*buf >= '0') && (*buf <= '9'))
   ++buf;
  if (*buf == 0)
  {
    return 0;
  }
  ++buf;
  if (!((*buf >= '0') && (*buf <= '9')))
  {
    printf("bad byte size option\n");
    return 0;
  }
  dcb->ByteSize = atoi(buf);

  while ((*buf >= '0') && (*buf <= '9'))
   ++buf;
  if (*buf == 0) return 0;
  ++buf;  // pass up ','

  switch(toupper(*buf++))
  {
    case 'N':  // none
      dcb->Parity = 0;
    break;
    case 'O':  // odd
      dcb->Parity = 1;
    break;
    case 'E':  // even
      dcb->Parity = 2;
    break;
    case 'M':  // mark
      dcb->Parity = 3;
    break;
    case 'S':  // space
      dcb->Parity = 4;
    break;
    default:
      printf("bad parity option\n");
      return 0;
  }

  while ((*buf != 0) && (*buf != ','))
    ++buf;
  if (*buf == 0) return 0;
  ++buf;
  if (!((*buf >= '0') && (*buf <= '9')))  return 0;
  if (*buf == 0) return 0;
  switch(*buf)
  {
    case '1':
      dcb->StopBits = 0;
    break;
    case '2':
      dcb->StopBits = 2;
    break;
    default:
      printf("bad stopbit option\n");
      return 0;
  }
  ++buf;

  while ((*buf != 0) && (*buf != ','))
    ++buf;
  if (*buf == 0) return 0;
  ++buf;  // pass up ','

  switch(toupper(*buf++))
  {
    case 'X': flow_control = 'X'; break;
    case 'R': flow_control = 'R'; break;
    case 'N': flow_control = 'N'; break;
    default:
      printf("bad flowcontrol option\n");
      return 0;
  }


  return 0;
}

/*--------------------------------------------------------------------
| f_set_com_options -
|---------------------------------------------------------------------*/
int f_set_com_options(int port,
                      DWORD BaudRate,
                      int Parity,
                      int StopBits,
                      int ByteSize,
                      int flow_ctrl)  // 0=none, 1 = rts,cts 2 = xon,xoff
{
 WORD val, bval, bsize, stops;

  switch (BaudRate)
  {
    case 110:   bval = 0; break;
    case 150:   bval = 1; break;
    case 300:   bval = 2; break;
    case 600:   bval = 3; break;
    case 1200:  bval = 4; break;
    case 2400:  bval = 5; break;
    case 4800:  bval = 6; break;
    case 9600:  bval = 7; break;
    case 19200: bval = 8; break;
    case 28800: bval = 0x80; break;
    case 38400: bval = 0x81; break;
    case 57600: bval = 0x82; break;
    case 76800: bval = 0x83; break;
    case 115200:bval = 0x84; break;
    case 230400:bval = 13; break;
    case 460800:bval = 14; break;
    case 921600:bval = 15; break;
    default  :  bval = 7; break;
  }

  switch (ByteSize)
  {
    case 5:
    case 6:
    case 7: bsize = 2; break;  // 7
    case 8: bsize = 3; break;  // 8
  }

  FOSSbaudExt(port, bval, bsize, Parity, StopBits);

  if (use_fossil)
  {
    val = 0;
    if (flow_ctrl & 1) val = 2;
    if (flow_ctrl & 2) val |= 9;

    FOSSflowCtrl(port, val);
  }
  return 0;
}

